#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import asyncio
import itertools
import json
import re
import sys
import threading
import time
from datetime import date, datetime, timezone
from typing import Any, Callable, Dict, List, Literal, Optional, Set, Tuple, Union
from urllib.parse import urljoin, urlparse

import numpy as np
import pandas as pd
from aiomysql.sa import create_engine as async_create_engine
from hsfs import client, feature, feature_group, serving_key
from hsfs.client import exceptions
from hsfs.client.exceptions import FeatureStoreException
from hsfs.constructor import serving_prepared_statement
from hsfs.core import feature_group_api, variable_api
from sqlalchemy import create_engine
from sqlalchemy.engine.url import make_url


FEATURE_STORE_NAME_SUFFIX = "_featurestore"


class FeatureStoreEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Dict[str, Any]:
        try:
            return o.to_dict()
        except AttributeError:
            return super().default(o)


def validate_feature(
    ft: Union[str, feature.Feature, Dict[str, Any]],
) -> feature.Feature:
    if isinstance(ft, feature.Feature):
        return ft
    elif isinstance(ft, str):
        return feature.Feature(ft)
    elif isinstance(ft, dict):
        return feature.Feature(**ft)


VALID_EMBEDDING_TYPE = {
    "array<int>",
    "array<bigint>",
    "array<float>",
    "array<double>",
}


def validate_embedding_feature_type(embedding_index, schema):
    if not embedding_index or not schema:
        return
    feature_type_map = dict([(feat.name, feat.type) for feat in schema])
    for embedding in embedding_index.get_embeddings():
        feature_type = feature_type_map.get(embedding.name)
        if feature_type not in VALID_EMBEDDING_TYPE:
            raise FeatureStoreException(
                f"Provide feature `{embedding.name}` has type `{feature_type}`, "
                f"but requires one of the following: {', '.join(VALID_EMBEDDING_TYPE)}"
            )


def parse_features(
    feature_names: Union[
        str, feature.Feature, List[Union[Dict[str, Any], str, feature.Feature]]
    ],
) -> List[feature.Feature]:
    if isinstance(feature_names, (str, feature.Feature)):
        return [validate_feature(feature_names)]
    elif isinstance(feature_names, list) and len(feature_names) > 0:
        return [validate_feature(feat) for feat in feature_names]
    else:
        return []


def autofix_feature_name(name: str) -> str:
    # replace spaces with underscores and enforce lower case
    return name.lower().replace(" ", "_")


def feature_group_name(
    feature_group: Union[
        feature_group.FeatureGroup,
        feature_group.ExternalFeatureGroup,
        feature_group.SpineGroup,
    ],
) -> str:
    return feature_group.name + "_" + str(feature_group.version)


def append_feature_store_suffix(name: str) -> str:
    name = name.lower()
    if name.endswith(FEATURE_STORE_NAME_SUFFIX):
        return name
    else:
        return name + FEATURE_STORE_NAME_SUFFIX


def strip_feature_store_suffix(name: str) -> str:
    name = name.lower()
    if name.endswith(FEATURE_STORE_NAME_SUFFIX):
        return name[: -1 * len(FEATURE_STORE_NAME_SUFFIX)]
    else:
        return name


def create_mysql_engine(
    online_conn: Any, external: bool, options: Optional[Dict[str, Any]] = None
) -> Any:
    online_options = online_conn.spark_options()
    # Here we are replacing the first part of the string returned by Hopsworks,
    # jdbc:mysql:// with the sqlalchemy one + username and password
    # useSSL and allowPublicKeyRetrieval are not valid properties for the pymysql driver
    # to use SSL we'll have to something like this:
    # ssl_args = {'ssl_ca': ca_path}
    # engine = create_engine("mysql+pymysql://<user>:<pass>@<addr>/<schema>", connect_args=ssl_args)
    if external:
        # This only works with external clients.
        # Hopsworks clients should use the storage connector
        host = get_host_name()
        online_options["url"] = re.sub(
            "/[0-9.]+:",
            "/{}:".format(host),
            online_options["url"],
        )

    sql_alchemy_conn_str = (
        online_options["url"]
        .replace(
            "jdbc:mysql://",
            "mysql+pymysql://"
            + online_options["user"]
            + ":"
            + online_options["password"]
            + "@",
        )
        .replace("useSSL=false&", "")
        .replace("?allowPublicKeyRetrieval=true", "")
    )
    if options is not None and not isinstance(options, dict):
        raise TypeError("`options` should be a `dict` type.")
    if not options:
        options = {"pool_recycle": 3600}
    elif "pool_recycle" not in options:
        options["pool_recycle"] = 3600
    # default connection pool size kept by engine is 5
    sql_alchemy_engine = create_engine(sql_alchemy_conn_str, **options)
    return sql_alchemy_engine


def get_host_name() -> str:
    host = variable_api.VariableApi().get_loadbalancer_external_domain()
    if host == "":
        # If the load balancer is not configured, then fall back to
        # use the MySQL node on the head node
        host = client.get_instance().host
    return host


def get_dataset_type(path: str) -> Literal["HIVEDB", "DATASET"]:
    if re.match(r"^(?:hdfs://|)/apps/hive/warehouse/*", path):
        return "HIVEDB"
    else:
        return "DATASET"


async def create_async_engine(
    online_conn: Any,
    external: bool,
    default_min_size: int,
    options: Optional[Dict[str, Any]] = None,
    hostname: Optional[str] = None,
) -> Any:
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError as er:
        raise RuntimeError(
            "Event loop is not running. Please invoke this co-routine from a running loop or provide an event loop."
        ) from er

    online_options = online_conn.spark_options()
    # read the keys user, password from online_conn as use them while creating the connection pool
    url = make_url(online_options["url"].replace("jdbc:", ""))
    if hostname is None:
        if external:
            hostname = get_host_name()
        else:
            hostname = url.host

    # create a aiomysql connection pool
    pool = await async_create_engine(
        host=hostname,
        port=3306,
        user=online_options["user"],
        password=online_options["password"],
        db=url.database,
        loop=loop,
        minsize=(
            options.get("minsize", default_min_size) if options else default_min_size
        ),
        maxsize=(
            options.get("maxsize", default_min_size) if options else default_min_size
        ),
        pool_recycle=(options.get("pool_recycle", -1) if options else -1),
    )
    return pool


def check_timestamp_format_from_date_string(input_date: str) -> Tuple[str, str]:
    date_format_patterns = {
        r"^([0-9]{4})([0-9]{2})([0-9]{2})$": "%Y%m%d",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})([0-9]{2})$": "%Y%m%d%H",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})$": "%Y%m%d%H%M",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})$": "%Y%m%d%H%M%S",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{3})$": "%Y%m%d%H%M%S%f",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})T([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{6})Z$": "ISO",
    }
    normalized_date = (
        input_date.replace("/", "")
        .replace("-", "")
        .replace(" ", "")
        .replace(":", "")
        .replace(".", "")
    )

    date_format = None
    for pattern in date_format_patterns:
        date_format_pattern = re.match(pattern, normalized_date)
        if date_format_pattern:
            date_format = date_format_patterns[pattern]
            break

    if date_format is None:
        raise ValueError(
            "Unable to identify format of the provided date value : " + input_date
        )

    return normalized_date, date_format


def get_timestamp_from_date_string(input_date: str) -> int:
    norm_input_date, date_format = check_timestamp_format_from_date_string(input_date)
    try:
        if date_format != "ISO":
            date_time = datetime.strptime(norm_input_date, date_format)
        else:
            date_time = datetime.fromisoformat(input_date[:-1])
    except ValueError as err:
        raise ValueError(
            "Unable to parse the normalized input date value : "
            + norm_input_date
            + " with format "
            + date_format
        ) from err
    if date_time.tzinfo is None:
        date_time = date_time.replace(tzinfo=timezone.utc)
    return int(float(date_time.timestamp()) * 1000)


def get_hudi_datestr_from_timestamp(timestamp: int) -> str:
    return datetime.utcfromtimestamp(timestamp / 1000).strftime("%Y%m%d%H%M%S%f")[:-3]


def get_delta_datestr_from_timestamp(timestamp: int) -> str:
    return datetime.utcfromtimestamp(timestamp / 1000).strftime("%Y-%m-%d %H:%M:%S.%f")[
        :-3
    ]


def convert_event_time_to_timestamp(
    event_time: Optional[
        Union[str, pd._libs.tslibs.timestamps.Timestamp, datetime, date, int]
    ],
) -> Optional[int]:
    if not event_time:
        return None
    if isinstance(event_time, str):
        return get_timestamp_from_date_string(event_time)
    elif isinstance(event_time, pd._libs.tslibs.timestamps.Timestamp):
        # convert to unix epoch time in milliseconds.
        event_time = event_time.to_pydatetime()
        # convert to unix epoch time in milliseconds.
        if event_time.tzinfo is None:
            event_time = event_time.replace(tzinfo=timezone.utc)
        return int(event_time.timestamp() * 1000)
    elif isinstance(event_time, datetime):
        # convert to unix epoch time in milliseconds.
        if event_time.tzinfo is None:
            event_time = event_time.replace(tzinfo=timezone.utc)
        return int(event_time.timestamp() * 1000)
    elif isinstance(event_time, date):
        # convert to unix epoch time in milliseconds.
        event_time = datetime(*event_time.timetuple()[:7])
        if event_time.tzinfo is None:
            event_time = event_time.replace(tzinfo=timezone.utc)
        return int(event_time.timestamp() * 1000)
    elif isinstance(event_time, int):
        if event_time == 0:
            raise ValueError("Event time should be greater than 0.")
        # jdbc supports timestamp precision up to second only.
        if len(str(event_time)) <= 10:
            event_time = event_time * 1000
        return event_time
    else:
        raise ValueError(
            "Given event time should be in `datetime`, `date`, `str` or `int` type"
        )


def setup_pydoop() -> None:
    # Import Pydoop only here, so it doesn't trigger if the execution environment
    # does not support Pydoop. E.g. Sagemaker
    from pydoop import hdfs

    # Create a subclass that replaces the check on the hdfs scheme to allow hopsfs as well.
    class _HopsFSPathSplitter(hdfs.path._HdfsPathSplitter):
        @classmethod
        def split(
            cls, hdfs_path: Optional[str], user: Optional[str]
        ) -> Tuple[str, int, str]:
            if not hdfs_path:
                cls.raise_bad_path(hdfs_path, "empty")
            scheme, netloc, path = cls.parse(hdfs_path)
            if not scheme:
                scheme = "file" if hdfs.fs.default_is_local() else "hdfs"
            if scheme == "hdfs" or scheme == "hopsfs":
                if not path:
                    cls.raise_bad_path(hdfs_path, "path part is empty")
                if ":" in path:
                    cls.raise_bad_path(hdfs_path, "':' not allowed outside netloc part")
                hostname, port = cls.split_netloc(netloc)
                if not path.startswith("/"):
                    path = "/user/%s/%s" % (user, path)
            elif scheme == "file":
                hostname, port, path = "", 0, netloc + path
            else:
                cls.raise_bad_path(hdfs_path, "unsupported scheme %r" % scheme)
            return hostname, port, path

    # Monkey patch the class to use the one defined above.
    hdfs.path._HdfsPathSplitter = _HopsFSPathSplitter


def get_hostname_replaced_url(sub_path: str) -> str:
    """
    construct and return an url with public hopsworks hostname and sub path
    :param self:
    :param sub_path: url sub-path after base url
    :return: href url
    """
    href = urljoin(client.get_instance()._base_url, sub_path)
    url_parsed = client.get_instance().replace_public_host(urlparse(href))
    return url_parsed.geturl()


def verify_attribute_key_names(
    feature_group_obj: Union[
        feature_group.FeatureGroup,
        feature_group.ExternalFeatureGroup,
        feature_group.SpineGroup,
    ],
    external_feature_group: bool = False,
) -> None:
    feature_names = set(feat.name for feat in feature_group_obj.features)
    if feature_group_obj.primary_key:
        diff = set(feature_group_obj.primary_key) - feature_names
        if diff:
            raise exceptions.FeatureStoreException(
                f"Provided primary key(s) {','.join(diff)} doesn't exist in feature dataframe"
            )

    if feature_group_obj.event_time:
        if feature_group_obj.event_time not in feature_names:
            raise exceptions.FeatureStoreException(
                f"Provided event_time feature {feature_group_obj.event_time} doesn't exist in feature dataframe"
            )

    if not external_feature_group:
        if feature_group_obj.partition_key:
            diff = set(feature_group_obj.partition_key) - feature_names
            if diff:
                raise exceptions.FeatureStoreException(
                    f"Provided partition key(s) {','.join(diff)} doesn't exist in feature dataframe"
                )

        if feature_group_obj.hudi_precombine_key:
            if feature_group_obj.hudi_precombine_key not in feature_names:
                raise exceptions.FeatureStoreException(
                    f"Provided hudi precombine key {feature_group_obj.hudi_precombine_key} "
                    f"doesn't exist in feature dataframe"
                )


def get_job_url(href: str) -> str:
    """Use the endpoint returned by the API to construct the UI url for jobs

    Args:
        href (str): the endpoint returned by the API
    """
    url = urlparse(href)
    url_splits = url.path.split("/")
    project_id = url_splits[4]
    job_name = url_splits[6]
    ui_url = url._replace(
        path="p/{}/jobs/named/{}/executions".format(project_id, job_name)
    )
    ui_url = client.get_instance().replace_public_host(ui_url)
    return ui_url.geturl()


def translate_legacy_spark_type(
    output_type: str,
) -> Literal[
    "STRING",
    "BINARY",
    "BYTE",
    "SHORT",
    "INT",
    "LONG",
    "FLOAT",
    "DOUBLE",
    "TIMESTAMP",
    "DATE",
    "BOOLEAN",
]:
    if output_type == "StringType()":
        return "STRING"
    elif output_type == "BinaryType()":
        return "BINARY"
    elif output_type == "ByteType()":
        return "BYTE"
    elif output_type == "ShortType()":
        return "SHORT"
    elif output_type == "IntegerType()":
        return "INT"
    elif output_type == "LongType()":
        return "LONG"
    elif output_type == "FloatType()":
        return "FLOAT"
    elif output_type == "DoubleType()":
        return "DOUBLE"
    elif output_type == "TimestampType()":
        return "TIMESTAMP"
    elif output_type == "DateType()":
        return "DATE"
    elif output_type == "BooleanType()":
        return "BOOLEAN"
    else:
        return "STRING"  # handle gracefully, and return STRING type, the default for spark udfs


def _loading_animation(message: str, stop_event: threading.Event) -> None:
    for char in itertools.cycle([".", "..", "...", ""]):
        if stop_event.is_set():
            break
        print(f"{message}{char}   ", end="\r")
        time.sleep(0.5)


def run_with_loading_animation(message: str, func: Callable, *args, **kwargs) -> Any:
    stop_event = threading.Event()
    t = threading.Thread(
        target=_loading_animation,
        args=(
            message,
            stop_event,
        ),
    )
    t.daemon = True
    t.start()
    start = time.time()
    end = None

    try:
        result = func(*args, **kwargs)
        end = time.time()
        return result
    finally:
        # Stop the animation and print the "Finished Querying" message
        stop_event.set()
        t.join()
        if not end:
            print(f"\rError: {message}           ", end="\n")
        else:
            print(f"\rFinished: {message} ({(end-start):.2f}s) ", end="\n")


def get_feature_group_url(feature_store_id: int, feature_group_id: int) -> str:
    sub_path = (
        "/p/"
        + str(client.get_instance()._project_id)
        + "/fs/"
        + str(feature_store_id)
        + "/fg/"
        + str(feature_group_id)
    )
    return get_hostname_replaced_url(sub_path)


def build_serving_keys_from_prepared_statements(
    prepared_statements: List[serving_prepared_statement.ServingPreparedStatement],
    feature_store_id: int,
    ignore_prefix: bool = False,
) -> Set[serving_key.ServingKey]:
    serving_keys = set()
    fg_api = feature_group_api.FeatureGroupApi()
    for statement in prepared_statements:
        fg = fg_api.get_by_id(feature_store_id, statement.feature_group_id)
        for param in statement.prepared_statement_parameters:
            serving_keys.add(
                serving_key.ServingKey(
                    feature_name=param.name,
                    join_index=statement.prepared_statement_index,
                    prefix=statement.prefix,
                    ignore_prefix=ignore_prefix,
                    feature_group=fg,
                )
            )
    return serving_keys


def is_runtime_notebook():
    if "ipykernel" in sys.modules:
        return True
    else:
        return False


class NpDatetimeEncoder(json.JSONEncoder):
    def default(self, obj):
        dtypes = (np.datetime64, np.complexfloating)
        if isinstance(obj, (datetime, date)):
            return convert_event_time_to_timestamp(obj)
        elif isinstance(obj, dtypes):
            return str(obj)
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            if any([np.issubdtype(obj.dtype, i) for i in dtypes]):
                return obj.astype(str).tolist()
            return obj.tolist()
        return super(NpDatetimeEncoder, self).default(obj)


class VersionWarning(Warning):
    pass


class JobWarning(Warning):
    pass


class StorageWarning(Warning):
    pass


class StatisticsWarning(Warning):
    pass


class ValidationWarning(Warning):
    pass


class FeatureGroupWarning(Warning):
    pass



#
#  Copyright 2021. Logical Clocks AB
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from __future__ import annotations

import json
from typing import Any, Dict, Optional

import humps
from hsfs import util


class PreparedStatementParameter:
    def __init__(
        self,
        name: Optional[str] = None,
        index: Optional[int] = None,
        type: Optional[str] = None,
        href: Optional[str] = None,
        **kwargs,
    ) -> None:
        self._name = name
        self._index = index

    @classmethod
    def from_response_json(
        cls, json_dict: Dict[str, Any]
    ) -> PreparedStatementParameter:
        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    def update_from_response_json(
        self, json_dict: Dict[str, Any]
    ) -> PreparedStatementParameter:
        json_decamelized = humps.decamelize(json_dict)
        self.__init__(**json_decamelized)
        return self

    def json(self) -> str:
        return json.dumps(self, cls=util.FeatureStoreEncoder)

    def to_dict(self) -> Dict[str, Any]:
        return {"name": self._name, "index": self._index}

    def __repr__(self) -> str:
        return humps.decamelize(json.dumps(self.to_dict(), sort_keys=True, indent=4))

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def index(self) -> Optional[int]:
        return self._index

    @name.setter
    def name(self, name: Optional[str]) -> None:
        self._name = name

    @index.setter
    def prepared_statement_index(self, index: Optional[int]) -> None:
        self._index = index



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsfs import transformation_function, transformation_function_attached


class TestTransformationFunctionAttached:
    def test_from_response_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["transformation_function_attached"]["get"]["response"]

        # Act
        tf_attached = transformation_function_attached.TransformationFunctionAttached.from_response_json(
            json
        )

        # Assert
        assert tf_attached.name == "test_name"
        assert isinstance(
            tf_attached.transformation_function,
            transformation_function.TransformationFunction,
        )

    def test_from_response_json_basic_info(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["transformation_function_attached"]["get_basic_info"][
            "response"
        ]

        # Act
        tf_attached = transformation_function_attached.TransformationFunctionAttached.from_response_json(
            json
        )

        # Assert
        assert tf_attached.name == "test_name"
        assert isinstance(
            tf_attached.transformation_function,
            transformation_function.TransformationFunction,
        )

    def test_from_response_json_list(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["transformation_function_attached"]["get_list"][
            "response"
        ]

        # Act
        tf_attached_list = transformation_function_attached.TransformationFunctionAttached.from_response_json(
            json
        )

        # Assert
        assert len(tf_attached_list) == 1
        tf_attached = tf_attached_list[0]
        assert tf_attached.name == "test_name"
        assert isinstance(
            tf_attached.transformation_function,
            transformation_function.TransformationFunction,
        )

    def test_from_response_json_list_empty(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["transformation_function_attached"]["get_list_empty"][
            "response"
        ]

        # Act
        tf_attached_list = transformation_function_attached.TransformationFunctionAttached.from_response_json(
            json
        )

        # Assert
        assert len(tf_attached_list) == 0



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsfs import (
    statistics_config,
    storage_connector,
    training_dataset,
    training_dataset_feature,
    training_dataset_split,
)


class TestTrainingDataset:
    def test_from_response_json(self, mocker, backend_fixtures):
        # Arrange
        mocker.patch("hsfs.client.get_instance")
        json = backend_fixtures["training_dataset"]["get"]["response"]

        # Act
        td_list = training_dataset.TrainingDataset.from_response_json(json)

        # Assert
        assert len(td_list) == 1
        td = td_list[0]
        assert td.id == 11
        assert td.name == "test_name"
        assert td.version == 1
        assert td.description == "test_description"
        assert td.data_format == "hudi"
        assert td._start_time == 1646438400000
        assert td._end_time == 1646697600000
        assert td.validation_size == 0.0
        assert td.test_size == 0.5
        assert td.train_start == 4
        assert td.train_end == 5
        assert td.validation_start == 6
        assert td.validation_end == 7
        assert td.test_start == 8
        assert td.test_end == 9
        assert td.coalesce is True
        assert td.seed == 123
        assert td.location == "test_location"
        assert td._from_query == "test_from_query"
        assert td._querydto == "test_querydto"
        assert td.feature_store_id == 22
        assert td.transformation_functions == "test_transformation_functions"
        assert td.train_split == "test_train_split"
        assert td.training_dataset_type == "HOPSFS_TRAINING_DATASET"
        assert isinstance(td.storage_connector, storage_connector.JdbcConnector)
        assert len(td._features) == 1
        assert isinstance(
            td._features[0], training_dataset_feature.TrainingDatasetFeature
        )
        assert len(td.splits) == 1
        assert isinstance(td.splits[0], training_dataset_split.TrainingDatasetSplit)
        assert isinstance(td.statistics_config, statistics_config.StatisticsConfig)
        assert td.label == ["test_name"]

    def test_from_response_json_basic_info(self, mocker, backend_fixtures):
        # Arrange
        mocker.patch("hsfs.client.get_instance")
        json = backend_fixtures["training_dataset"]["get_basic_info"]["response"]

        # Act
        td_list = training_dataset.TrainingDataset.from_response_json(json)

        # Assert
        assert len(td_list) == 1
        td = td_list[0]
        assert td.id is None
        assert td.name == "test_name"
        assert td.version == 1
        assert td.description is None
        assert td.data_format == "hudi"
        assert td._start_time is None
        assert td._end_time is None
        assert td.validation_size is None
        assert td.test_size is None
        assert td.train_start is None
        assert td.train_end is None
        assert td.validation_start is None
        assert td.validation_end is None
        assert td.test_start is None
        assert td.test_end is None
        assert td.coalesce is False
        assert td.seed is None
        assert td.location == ""
        assert td._from_query is None
        assert td._querydto is None
        assert td.feature_store_id == 22
        assert td.transformation_functions is None
        assert td.train_split is None
        assert td.training_dataset_type is None
        assert isinstance(td.storage_connector, storage_connector.JdbcConnector)
        assert len(td._features) == 0
        assert len(td.splits) == 0
        assert isinstance(td.statistics_config, statistics_config.StatisticsConfig)
        assert td.label == []

    def test_from_response_json_empty(self, mocker, backend_fixtures):
        # Arrange
        mocker.patch("hsfs.client.get_instance")
        json = backend_fixtures["training_dataset"]["get_empty"]["response"]

        # Act
        td_list = training_dataset.TrainingDataset.from_response_json(json)

        # Assert
        assert len(td_list) == 0



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsfs import split_statistics, statistics
from hsfs.core.feature_descriptive_statistics import FeatureDescriptiveStatistics


class TestStatistics:
    def test_from_response_json_no_items(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["statistics"]["get_with_no_items"]["response"]

        # Act
        s = statistics.Statistics.from_response_json(json)

        # Assert
        assert s is None

    def test_from_response_json_no_filtered_items(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["statistics"]["get_with_no_filtered_items"]["response"]

        # Act
        s = statistics.Statistics.from_response_json(json)

        # Assert
        assert s is None

    def test_from_response_json_without_items_field(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["statistics"]["get_without_items_field"]["response"]

        stats = FeatureDescriptiveStatistics.from_response_json(
            backend_fixtures["feature_descriptive_statistics"][
                "get_fractional_feature_statistics"
            ]["response"]
        )

        # Act
        s = statistics.Statistics.from_response_json(json)

        # Assert
        assert isinstance(s, statistics.Statistics)
        assert s.computation_time == "test_computation_time"
        assert s.window_end_commit_time == 11
        assert s.feature_descriptive_statistics is not None
        assert len(s.feature_descriptive_statistics) == 1
        assert isinstance(
            s.feature_descriptive_statistics[0], FeatureDescriptiveStatistics
        )
        assert s.feature_descriptive_statistics[0].id == stats.id
        assert s.split_statistics is None
        assert not s.before_transformation

    def test_from_response_json_single(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["statistics"]["get"]["response"]

        stats = FeatureDescriptiveStatistics.from_response_json(
            backend_fixtures["feature_descriptive_statistics"][
                "get_fractional_feature_statistics"
            ]["response"]
        )

        # Act
        s_lst = statistics.Statistics.from_response_json(json)

        # Assert
        assert isinstance(s_lst, list)
        assert len(s_lst) == 1
        s = s_lst[0]
        assert s.computation_time == "test_computation_time"
        assert s.window_end_commit_time == 11
        assert s.feature_descriptive_statistics is not None
        assert len(s.feature_descriptive_statistics) == 1
        assert isinstance(
            s.feature_descriptive_statistics[0], FeatureDescriptiveStatistics
        )
        assert s.feature_descriptive_statistics[0].id == stats.id
        assert s.split_statistics is None
        assert not s.before_transformation

    def test_from_response_json_multiple(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["statistics"]["get"]["response"]

        stats = FeatureDescriptiveStatistics.from_response_json(
            backend_fixtures["feature_descriptive_statistics"][
                "get_fractional_feature_statistics"
            ]["response"]
        )

        # Act
        s_lst = statistics.Statistics.from_response_json(json)

        # Assert
        assert isinstance(s_lst, list)
        assert len(s_lst) == 1
        s = s_lst[0]
        assert s.computation_time == "test_computation_time"
        assert s.window_end_commit_time == 11
        assert s.feature_descriptive_statistics is not None
        assert len(s.feature_descriptive_statistics) == 1
        assert isinstance(
            s.feature_descriptive_statistics[0], FeatureDescriptiveStatistics
        )
        assert s.feature_descriptive_statistics[0].id == stats.id
        assert s.split_statistics is None
        assert not s.before_transformation

    def test_from_response_json_with_split_statistics(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["statistics"]["get_with_split_statistics"]["response"]

        stats = FeatureDescriptiveStatistics.from_response_json(
            backend_fixtures["feature_descriptive_statistics"][
                "get_fractional_feature_statistics"
            ]["response"]
        )

        # Act
        s_lst = statistics.Statistics.from_response_json(json)
        s = s_lst[0]

        # Assert
        assert s.computation_time == "test_computation_time"
        assert s.window_end_commit_time == 11
        assert s.feature_descriptive_statistics is None
        assert s.split_statistics is not None
        assert len(s.split_statistics) == 1
        assert isinstance(s.split_statistics[0], split_statistics.SplitStatistics)
        assert len(s.split_statistics[0].feature_descriptive_statistics) == 1
        assert isinstance(
            s.split_statistics[0].feature_descriptive_statistics[0],
            FeatureDescriptiveStatistics,
        )
        assert s.split_statistics[0].feature_descriptive_statistics[0].id == stats.id
        assert s.before_transformation is False

    def test_from_response_json_with_empty_data(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["statistics"]["get_with_empty_data"]["response"]

        stats = FeatureDescriptiveStatistics.from_response_json(
            backend_fixtures["feature_descriptive_statistics"]["get_for_empty_data"][
                "response"
            ]
        )

        # Act
        s_lst = statistics.Statistics.from_response_json(json)
        s = s_lst[0]

        # Assert
        assert s.computation_time == "test_computation_time"
        assert s.window_end_commit_time is None
        assert s.feature_descriptive_statistics is not None
        assert len(s.feature_descriptive_statistics) == 1
        assert isinstance(
            s.feature_descriptive_statistics[0], FeatureDescriptiveStatistics
        )
        assert s.feature_descriptive_statistics[0].id == stats.id
        assert s.feature_descriptive_statistics[0].count == 0
        assert s.split_statistics is None
        assert s.before_transformation is False

    def test_from_response_json_before_transformation_functions(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["statistics"]["get_before_transformation_functions"][
            "response"
        ]

        stats = FeatureDescriptiveStatistics.from_response_json(
            backend_fixtures["feature_descriptive_statistics"][
                "get_before_transformation_functions"
            ]["response"]
        )

        # Act
        s_lst = statistics.Statistics.from_response_json(json)
        s = s_lst[0]

        # Assert
        assert s.computation_time == "test_computation_time"
        assert s.window_end_commit_time is None
        assert s.feature_descriptive_statistics is not None
        assert len(s.feature_descriptive_statistics) == 1
        assert isinstance(
            s.feature_descriptive_statistics[0], FeatureDescriptiveStatistics
        )
        assert s.feature_descriptive_statistics[0].id == stats.id
        assert s.feature_descriptive_statistics[0].count is None
        assert s.feature_descriptive_statistics[0].extended_statistics is not None
        assert (
            "unique_values" in s.feature_descriptive_statistics[0].extended_statistics
        )
        assert s.split_statistics is None
        assert s.before_transformation is True

    def test_from_response_json_basic_info(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["statistics"]["get_basic_info"]["response"]

        stats = FeatureDescriptiveStatistics.from_response_json(
            backend_fixtures["feature_descriptive_statistics"][
                "get_fractional_feature_statistics"
            ]["response"]
        )

        # Act
        s_lst = statistics.Statistics.from_response_json(json)
        s = s_lst[0]

        # Assert
        assert s.computation_time == "test_computation_time"
        assert s.window_end_commit_time is None
        assert s.feature_descriptive_statistics is not None
        assert len(s.feature_descriptive_statistics) == 1
        assert isinstance(
            s.feature_descriptive_statistics[0], FeatureDescriptiveStatistics
        )
        assert s.feature_descriptive_statistics[0].id == stats.id
        assert s.split_statistics is None
        assert s.before_transformation is False

    def test_from_response_json_empty(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["statistics"]["get_empty"]["response"]

        # Act
        s = statistics.Statistics.from_response_json(json)

        # Assert
        assert s is None



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import Any, Dict

from hsfs import client, storage_connector


class StorageConnectorApi:
    def _get(self, feature_store_id: int, name: str) -> Dict[str, Any]:
        """Returning response dict instead of initialized object."""
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "storageconnectors",
            name,
        ]
        query_params = {"temporaryCredentials": True}
        return _client._send_request("GET", path_params, query_params=query_params)

    def get(
        self, feature_store_id: int, name: str
    ) -> storage_connector.StorageConnector:
        """Get storage connector with name.

        :param feature_store_id: feature store id
        :type feature_store_id: int
        :param name: name of the storage connector
        :type name: str
        :return: the storage connector
        :rtype: StorageConnector
        """
        return storage_connector.StorageConnector.from_response_json(
            self._get(feature_store_id, name)
        )

    def refetch(
        self, storage_connector_instance: storage_connector.StorageConnector
    ) -> storage_connector.StorageConnector:
        """
        Refetches the storage connector from Hopsworks, in order to update temporary
        credentials.
        """
        return storage_connector_instance.update_from_response_json(
            self._get(
                storage_connector_instance._featurestore_id,
                storage_connector_instance.name,
            )
        )

    def get_online_connector(
        self, feature_store_id: int
    ) -> storage_connector.OnlineStorageConnector:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "storageconnectors",
            "onlinefeaturestore",
        ]

        return storage_connector.StorageConnector.from_response_json(
            _client._send_request("GET", path_params)
        )

    def get_kafka_connector(
        self, feature_store_id: int, external: bool = False
    ) -> storage_connector.KafkaConnector:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "storageconnectors",
            "kafka_connector",
            "byok",
        ]
        query_params = {"external": external}

        return storage_connector.StorageConnector.from_response_json(
            _client._send_request("GET", path_params, query_params=query_params)
        )

    def get_feature_groups_provenance(
        self, storage_connector_instance
    ):
        """Get the generated feature groups using this storage connector, based on explicit
        provenance. These feature groups can be accessible or inaccessible. Explicit
        provenance does not track deleted generated feature group links, so deleted
        will always be empty.
        For inaccessible feature groups, only a minimal information is returned.

        # Arguments
            storage_connector_instance: Metadata object of storage connector.

        # Returns
            `ExplicitProvenance.Links`: the feature groups generated using this
            storage connector
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            storage_connector_instance._featurestore_id,
            "storageconnectors",
            storage_connector_instance.name,
            "provenance",
            "links",
        ]
        query_params = {
            "expand": "provenance_artifacts",
            "upstreamLvls": 0,
            "downstreamLvls": 1,
        }
        links_json = _client._send_request("GET", path_params, query_params)
        from hsfs.core import explicit_provenance
        return explicit_provenance.Links.from_response_json(
            links_json,
            explicit_provenance.Links.Direction.DOWNSTREAM,
            explicit_provenance.Links.Type.FEATURE_GROUP,
        )



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import humps
from hsfs.core.job import Job


class IngestionJob:
    def __init__(
        self,
        data_path,
        job,
        href=None,
        expand=None,
        items=None,
        count=None,
        type=None,
        **kwargs,
    ):
        self._data_path = data_path
        self._job = Job.from_response_json(job)

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    @property
    def data_path(self):
        return self._data_path

    @property
    def job(self):
        return self._job



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsfs import split_statistics
from hsfs.core.feature_descriptive_statistics import FeatureDescriptiveStatistics


class TestSplitStatistics:
    def test_from_response_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["split_statistics"]["get"]["response"]

        statistics = FeatureDescriptiveStatistics.from_response_json(
            backend_fixtures["feature_descriptive_statistics"][
                "get_fractional_feature_statistics"
            ]["response"]
        )

        # Act
        ss = split_statistics.SplitStatistics.from_response_json(json)

        # Assert
        assert ss.name == "test_name"
        assert len(ss.feature_descriptive_statistics) == 1
        assert isinstance(
            ss.feature_descriptive_statistics[0], FeatureDescriptiveStatistics
        )
        assert ss.feature_descriptive_statistics[0].id == statistics.id

    def test_from_response_json_basic_info(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["split_statistics"]["get_basic_info"]["response"]

        statistics = FeatureDescriptiveStatistics.from_response_json(
            backend_fixtures["feature_descriptive_statistics"][
                "get_fractional_feature_statistics"
            ]["response"]
        )

        # Act
        ss = split_statistics.SplitStatistics.from_response_json(json)

        # Assert
        assert ss.name == "test_name"
        assert len(ss.feature_descriptive_statistics) == 1
        assert isinstance(
            ss.feature_descriptive_statistics[0], FeatureDescriptiveStatistics
        )
        assert ss.feature_descriptive_statistics[0].id == statistics.id



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import warnings

from hsfs import feature_view, training_dataset_feature, transformation_function
from hsfs.constructor import fs_query, query
from hsfs.feature_store import FeatureStore


class TestFeatureView:
    def test_from_response_json(self, mocker, backend_fixtures):
        # Arrange
        mocker.patch.object(
            FeatureStore,
            "project_id",
            return_value=99,
        )
        mocker.patch("hsfs.client.get_instance")
        mocker.patch("hsfs.engine.get_type")
        mocker.patch("hsfs.core.feature_store_api.FeatureStoreApi.get")
        json = backend_fixtures["feature_view"]["get"]["response"]

        # Act
        fv = feature_view.FeatureView.from_response_json(json)

        # Assert
        assert fv.name == "test_name"
        assert fv.id == 11
        assert isinstance(fv.query, query.Query)
        assert fv.featurestore_id == 5
        assert fv.version == 1
        assert fv.description == "test_description"
        assert fv.labels == ["intt"]
        assert fv.transformation_functions == {}
        assert len(fv.schema) == 2
        assert isinstance(fv.schema[0], training_dataset_feature.TrainingDatasetFeature)

    def test_from_response_json_basic_info(self, mocker, backend_fixtures):
        # Arrange
        mocker.patch("hsfs.engine.get_type")
        mocker.patch("hsfs.client.get_instance")
        json = backend_fixtures["feature_view"]["get_basic_info"]["response"]

        # Act
        fv = feature_view.FeatureView.from_response_json(json)

        # Assert
        assert fv.name == "test_name"
        assert fv.id is None
        assert isinstance(fv.query, query.Query)
        assert fv.featurestore_id == 5
        assert fv.version is None
        assert fv.description is None
        assert fv.labels == []
        assert fv.transformation_functions == {}
        assert len(fv.schema) == 0
        assert fv.query._left_feature_group.deprecated is False

    def test_from_response_json_basic_info_deprecated(self, mocker, backend_fixtures):
        # Arrange
        mocker.patch("hsfs.engine.get_type")
        mocker.patch("hsfs.client.get_instance")
        json = backend_fixtures["feature_view"]["get_basic_info_deprecated"]["response"]

        # Act
        with warnings.catch_warnings(record=True) as warning_record:
            fv = feature_view.FeatureView.from_response_json(json)

        # Assert
        assert fv.name == "test_name"
        assert fv.id is None
        assert isinstance(fv.query, query.Query)
        assert fv.featurestore_id == 5
        assert fv.version is None
        assert fv.description is None
        assert fv.labels == []
        assert fv.transformation_functions == {}
        assert len(fv.schema) == 0
        assert fv.query._left_feature_group.deprecated is True
        assert len(warning_record) == 1
        assert str(warning_record[0].message) == (
            f"Feature Group `{fv.query._left_feature_group.name}`, version `{fv.query._left_feature_group.version}` is deprecated"
        )

    def test_transformation_function_instances(self, mocker, backend_fixtures):
        # Arrange
        feature_store_id = 99
        mocker.patch("hsfs.core.feature_view_engine.FeatureViewEngine")
        json = backend_fixtures["fs_query"]["get"]["response"]

        # Act
        q = fs_query.FsQuery.from_response_json(json)

        def testFunction():
            print("Test")

        tf = transformation_function.TransformationFunction(
            feature_store_id,
            transformation_fn=testFunction,
            builtin_source_code="",
            output_type="str",
        )

        transformation_fn_dict = dict()
        transformation_fn_dict["tf_name"] = tf
        transformation_fn_dict["tf1_name"] = tf

        fv = feature_view.FeatureView(
            featurestore_id=feature_store_id,
            name="test_fv",
            version=1,
            query=q,
            transformation_functions=transformation_fn_dict,
        )

        updated_transformation_fn_dict = fv.transformation_functions

        assert (
            updated_transformation_fn_dict["tf_name"]
            != updated_transformation_fn_dict["tf1_name"]
        )



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


import great_expectations as ge
from hsfs import ge_expectation


class TestGeExpectation:
    def test_from_response_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["ge_expectation"]["get"]["response"]

        # Act
        expect = ge_expectation.GeExpectation.from_response_json(json)

        # Assert
        print(expect.meta)
        assert expect.expectation_type == "1"
        assert expect.kwargs == {"kwargs_key": "kwargs_value"}
        assert expect.meta["meta_key"] == "meta_value"
        assert expect.meta["expectationId"] == 32
        assert expect.id == 32

    def test_from_response_json_list(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["ge_expectation"]["get_list"]["response"]

        # Act
        ge_list = ge_expectation.GeExpectation.from_response_json(json)

        # Assert
        assert len(ge_list) == 1
        expect = ge_list[0]
        print(expect.meta)
        assert expect.expectation_type == "1"
        assert expect.kwargs == {"kwargs_key": "kwargs_value"}
        assert expect.meta["meta_key"] == "meta_value"
        assert expect.meta["expectationId"] == 32
        assert expect.id == 32

    def test_from_response_json_list_empty(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["ge_expectation"]["get_list_empty"]["response"]

        # Act
        ge_list = ge_expectation.GeExpectation.from_response_json(json)

        # Assert
        assert len(ge_list) == 0

    def test_from_ge_object(self):
        # Arrange
        expectationId = 32
        expectation_type = "expect_column_min_to_be_between"
        kwargs = {"kwargs_key": "kwargs_value"}
        meta = {"meta_key": "meta_value", "expectationId": expectationId}
        ge_object = ge.core.ExpectationConfiguration(
            expectation_type=expectation_type,
            kwargs=kwargs,
            meta=meta,
        )

        # Act
        expect = ge_expectation.GeExpectation.from_ge_type(ge_object)

        # Assert
        assert expect.id == 32
        assert expect.expectation_type == expectation_type
        assert expect.kwargs == kwargs
        assert expect.meta == meta



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import List

from hsfs import client
from hsfs.ge_expectation import GeExpectation


class ExpectationApi:
    def __init__(
        self, feature_store_id: int, feature_group_id: int, expectation_suite_id: int
    ):
        """Expectation Suite endpoints for the featuregroup resource.
        :param feature_store_id: id of the respective Feature Store
        :type feature_store_id: int
        :param feature_group_id: id of the respective Feature Group
        :type feature_group_id: int
        :param expectation_suite_id: id of the respective Expectation Suite
        :type expectation_suite_id: int
        """
        self._feature_store_id = feature_store_id
        self._feature_group_id = feature_group_id
        self._expectation_suite_id = expectation_suite_id

    def create(self, expectation: GeExpectation) -> GeExpectation:
        """Create an expectation suite attached to a Feature Group.
        :param expectation: Expectation object to be appended to an Expectation Suite
        :type expectation: `GeExpectation`

        :return: expectation
        :rtype: `GeExpectation`
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "expectationsuite",
            self._expectation_suite_id,
            "expectations",
        ]

        headers = {"content-type": "application/json"}
        payload = expectation.json()
        return GeExpectation.from_response_json(
            _client._send_request("POST", path_params, headers=headers, data=payload)
        )

    def update(self, expectation: GeExpectation) -> GeExpectation:
        """Update an Expectation of an Expectation Suite attached to a Feature Group.
        :param expectation: Expectation object to be appended to an Expectation Suite
        :type expectation: `GeExpectation`

        :return: expectation
        :rtype: `GeExpectation`
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "expectationsuite",
            self._expectation_suite_id,
            "expectations",
            expectation.id,
        ]

        headers = {"content-type": "application/json"}
        payload = expectation.json()
        return GeExpectation.from_response_json(
            _client._send_request("PUT", path_params, headers=headers, data=payload)
        )

    def delete(self, expectation_id: int) -> None:
        """Delete the Expectation with expectation_id from the Expectation Suite attached to a Feature Group.
        :param expectation_id: id of the Expectation to delete
        :type expectation_id: `int`
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "expectationsuite",
            self._expectation_suite_id,
            "expectations",
            expectation_id,
        ]

        _client._send_request("DELETE", path_params)

    def get(self, expectation_id: int) -> GeExpectation:
        """Get an expectation attached to a feature group.

        :return: expectation
        :rtype: `GeExpectation`
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "expectationsuite",
            self._expectation_suite_id,
            "expectations",
            expectation_id,
        ]

        return GeExpectation.from_response_json(
            _client._send_request("GET", path_params)
        )

    def get_expectations_by_suite_id(self) -> List[GeExpectation]:
        """Get an expectation attached to a feature group.

        :return: expectation
        :rtype: `GeExpectation`
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "expectationsuite",
            self._expectation_suite_id,
            "expectations",
        ]

        return GeExpectation.from_response_json(
            _client._send_request("GET", path_params)
        )



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsfs import ge_validation_result, validation_report


class TestValidationReport:
    def test_from_response_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["validation_report"]["get"]["response"]

        # Act
        vr = validation_report.ValidationReport.from_response_json(json)

        # Assert
        assert vr.id == 11
        assert vr.success is True
        assert vr._full_report_path == "test_full_report_path"
        assert vr._validation_time == "test_validation_time"
        assert vr._featurestore_id == "test_featurestore_id"
        assert vr._featuregroup_id == "test_featuregroup_id"
        assert vr.ingestion_result == "INGESTED"
        assert len(vr.results) == 1
        assert isinstance(vr.results[0], ge_validation_result.ValidationResult)
        assert vr.meta == {"meta_key": "meta_value"}
        assert vr.statistics == {"statistics_key": "statistics_value"}
        assert vr.evaluation_parameters == {
            "evaluation_parameters_key": "evaluation_parameters_value"
        }

    def test_from_response_json_basic_info(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["validation_report"]["get_basic_info"]["response"]

        # Act
        vr = validation_report.ValidationReport.from_response_json(json)

        # Assert
        assert vr.id is None
        assert vr.success is True
        assert vr._full_report_path is None
        assert vr._validation_time is None
        assert vr._featurestore_id is None
        assert vr._featuregroup_id is None
        assert vr.ingestion_result == "UNKNOWN"
        assert len(vr.results) == 1
        assert isinstance(vr.results[0], ge_validation_result.ValidationResult)
        assert vr.meta == {"meta_key": "meta_value"}
        assert vr.statistics == {"statistics_key": "statistics_value"}
        assert vr.evaluation_parameters is None

    def test_from_response_json_list(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["validation_report"]["get_list"]["response"]

        # Act
        vr_list = validation_report.ValidationReport.from_response_json(json)

        # Assert
        assert len(vr_list) == 1
        vr = vr_list[0]
        assert vr.id == 11
        assert vr.success is True
        assert vr._full_report_path == "test_full_report_path"
        assert vr._validation_time == "test_validation_time"
        assert vr._featurestore_id == "test_featurestore_id"
        assert vr._featuregroup_id == "test_featuregroup_id"
        assert vr.ingestion_result == "test_ingestion_result"
        assert len(vr.results) == 1
        assert isinstance(vr.results[0], ge_validation_result.ValidationResult)
        assert vr.meta == {"meta_key": "meta_value"}
        assert vr.statistics == {"statistics_key": "statistics_value"}
        assert vr.evaluation_parameters == {
            "evaluation_parameters_key": "evaluation_parameters_value"
        }

    def test_from_response_json_list_empty(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["validation_report"]["get_list_empty"]["response"]

        # Act
        vr_list = validation_report.ValidationReport.from_response_json(json)

        # Assert
        assert len(vr_list) == 0



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import Any, Dict, Union

from hsfs import client
from hsfs.core import (
    execution,
    ingestion_job_conf,
    job,
    job_configuration,
    job_schedule,
)


class JobApi:
    def create(
        self,
        name: str,
        job_conf: Union[
            job_configuration.JobConfiguration, ingestion_job_conf.IngestionJobConf
        ],
    ) -> job.Job:
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", name]

        headers = {"content-type": "application/json"}
        return job.Job.from_response_json(
            _client._send_request(
                "PUT", path_params, headers=headers, data=job_conf.json()
            )
        )

    def launch(self, name: str, args: str = None) -> None:
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", name, "executions"]

        _client._send_request("POST", path_params, data=args)

    def get(self, name: str) -> job.Job:
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", name]

        return job.Job.from_response_json(_client._send_request("GET", path_params))

    def last_execution(self, job: job.Job) -> execution.Execution:
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", job.name, "executions"]

        query_params = {"limit": 1, "sort_by": "submissiontime:desc"}

        headers = {"content-type": "application/json"}
        return execution.Execution.from_response_json(
            _client._send_request(
                "GET", path_params, headers=headers, query_params=query_params
            )
        )

    def create_or_update_schedule_job(
        self, name: str, schedule_config: Dict[str, Any]
    ) -> job_schedule.JobSchedule:
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", name, "schedule", "v2"]
        headers = {"content-type": "application/json"}
        method = "PUT" if schedule_config["id"] else "POST"

        return job_schedule.JobSchedule.from_response_json(
            _client._send_request(
                method, path_params, headers=headers, data=json.dumps(schedule_config)
            )
        )

    def delete_schedule_job(self, name: str) -> None:
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", name, "schedule", "v2"]

        return _client._send_request(
            "DELETE",
            path_params,
        )



#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import Any, Dict, List, Optional

import humps


class Inode:
    def __init__(
        self,
        attributes: Dict[str, Any],
        href: Optional[str] = None,
        zip_state: Optional[str] = None,
        tags: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        self._path = attributes["path"]

    @classmethod
    def from_response_json(cls, json_dict: Dict[str, Any]) -> List[Inode]:
        json_decamelized = humps.decamelize(json_dict)["items"]
        for inode in json_decamelized:
            _ = inode.pop("type", None)
        return [cls(**inode) for inode in json_decamelized]

    @property
    def path(self) -> str:
        return self._path



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from datetime import date, datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import humps
from hsfs import util
from hsfs.client.exceptions import FeatureStoreException
from hsfs.core import (
    feature_monitoring_config_engine,
    feature_monitoring_result_engine,
    job_api,
    monitoring_window_config_engine,
)
from hsfs.core import monitoring_window_config as mwc
from hsfs.core.feature_monitoring_result import FeatureMonitoringResult
from hsfs.core.job_schedule import JobSchedule


MAX_LENGTH_NAME = 63
MAX_LENGTH_DESCRIPTION = 2000


class FeatureMonitoringType(str, Enum):
    STATISTICS_COMPUTATION = "STATISTICS_COMPUTATION"  # only stats computation
    STATISTICS_COMPARISON = "STATISTICS_COMPARISON"  # stats computation and comparison
    PROBABILITY_DENSITY_FUNCTION = "PROBABILITY_DENSITY_FUNCTION"  # data distributions

    @classmethod
    def list_str(cls) -> List[str]:
        return list(map(lambda c: c.value, cls))

    @classmethod
    def list(cls) -> List["FeatureMonitoringType"]:
        return list(map(lambda c: c, cls))

    @classmethod
    def from_str(cls, value: str) -> "FeatureMonitoringType":
        if value in cls.list_str():
            return cls(value)
        else:
            raise ValueError(
                f"Invalid value {value} for FeatureMonitoringType, allowed values are {cls.list_str()}"
            )

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        if isinstance(other, self.__class__):
            return self is other
        return False

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value


class FeatureMonitoringConfig:
    def __init__(
        self,
        feature_store_id: int,
        name: str,
        feature_name: Optional[str] = None,
        feature_monitoring_type: Union[
            FeatureMonitoringType, str
        ] = FeatureMonitoringType.STATISTICS_COMPUTATION,
        job_name: Optional[str] = None,
        detection_window_config: Optional[
            Union[mwc.MonitoringWindowConfig, Dict[str, Any]]
        ] = None,
        reference_window_config: Optional[
            Union[mwc.MonitoringWindowConfig, Dict[str, Any]]
        ] = None,
        statistics_comparison_config: Optional[Dict[str, Any]] = None,
        job_schedule: Optional[Union[Dict[str, Any], JobSchedule]] = None,
        description: Optional[str] = None,
        id: Optional[int] = None,
        feature_group_id: Optional[int] = None,
        feature_view_name: Optional[str] = None,
        feature_view_version: Optional[int] = None,
        href: Optional[str] = None,
        **kwargs,
    ):
        self.name = name
        self._id = id
        self._href = href
        self.description = description
        self._feature_name = feature_name
        self._feature_store_id = feature_store_id
        self._feature_group_id = feature_group_id
        self._feature_view_name = feature_view_name
        self._feature_view_version = feature_view_version
        self._job_name = job_name
        self._feature_monitoring_type = FeatureMonitoringType(feature_monitoring_type)

        self._feature_monitoring_config_engine = (
            feature_monitoring_config_engine.FeatureMonitoringConfigEngine(
                feature_store_id=feature_store_id,
                feature_group_id=feature_group_id,
                feature_view_name=feature_view_name,
                feature_view_version=feature_view_version,
            )
        )
        self._monitoring_window_config_engine = (
            monitoring_window_config_engine.MonitoringWindowConfigEngine()
        )
        self._feature_monitoring_result_engine = (
            feature_monitoring_result_engine.FeatureMonitoringResultEngine(
                feature_store_id=feature_store_id,
                feature_group_id=feature_group_id,
                feature_view_name=feature_view_name,
                feature_view_version=feature_view_version,
            )
        )
        self._job_api = job_api.JobApi()

        self.detection_window_config = detection_window_config
        self.reference_window_config = reference_window_config
        self.statistics_comparison_config = statistics_comparison_config
        self.job_schedule = job_schedule

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if "count" in json_decamelized:
            if json_decamelized["count"] == 0:
                return []
            return [cls(**config) for config in json_decamelized["items"]]
        else:
            return cls(**json_decamelized)

    def to_dict(self):
        detection_window_config = (
            self._detection_window_config.to_dict()
            if self._detection_window_config is not None
            else None
        )
        reference_window_config = (
            self._reference_window_config.to_dict()
            if self._reference_window_config is not None
            else None
        )
        if isinstance(self._statistics_comparison_config, dict):
            statistics_comparison_config = {
                "id": self._statistics_comparison_config.get("id", None),
                "threshold": self._statistics_comparison_config.get("threshold", 0.0),
                "metric": self._statistics_comparison_config.get("metric", "MEAN"),
                "strict": self._statistics_comparison_config.get("strict", False),
                "relative": self._statistics_comparison_config.get("relative", False),
            }
        else:
            statistics_comparison_config = None
        if isinstance(self._job_schedule, JobSchedule):
            job_schedule = self._job_schedule.to_dict()
        else:
            job_schedule = None

        the_dict = {
            "id": self._id,
            "featureStoreId": self._feature_store_id,
            "featureName": self._feature_name,
            "name": self._name,
            "description": self._description,
            "jobName": self._job_name,
            "featureMonitoringType": self._feature_monitoring_type,
            "jobSchedule": job_schedule,
            "detectionWindowConfig": detection_window_config,
        }

        if self._feature_group_id is not None:
            the_dict["featureGroupId"] = self._feature_group_id
        elif (
            self._feature_view_name is not None
            and self._feature_view_version is not None
        ):
            the_dict["featureViewName"] = self._feature_view_name
            the_dict["featureViewVersion"] = self._feature_view_version

        if (
            self._feature_monitoring_type
            == FeatureMonitoringType.STATISTICS_COMPUTATION
        ):
            return the_dict

        the_dict["referenceWindowConfig"] = reference_window_config
        the_dict["statisticsComparisonConfig"] = statistics_comparison_config

        return the_dict

    def json(self) -> str:
        return json.dumps(self, cls=util.FeatureStoreEncoder)

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"FeatureMonitoringConfig({self._name!r}, {self._feature_monitoring_type!r})"

    def with_detection_window(
        self,
        time_offset: Optional[str] = None,
        window_length: Optional[str] = None,
        row_percentage: Optional[float] = None,
    ) -> "FeatureMonitoringConfig":
        """Sets the detection window of data to compute statistics on.
        !!! example
            ```python
            # Fetch your feature group or feature view
            fg = fs.get_feature_group(name="my_feature_group", version=1)
            # Compute statistics on a regular basis
            fg.create_statistics_monitoring(
                name="regular_stats",
                cron_expression="0 0 12 ? * * *",
            ).with_detection_window(
                time_offset="1d",
                window_length="1d",
                row_percentage=0.1,
            ).save()
            # Compute and compare statistics
            fg.create_feature_monitoring(
                name="regular_stats",
                feature_name="my_feature",
                cron_expression="0 0 12 ? * * *",
            ).with_detection_window(
                time_offset="1d",
                window_length="1d",
                row_percentage=0.1,
            ).with_reference_window(...).compare_on(...).save()
            ```
        # Arguments
            time_offset: The time offset from the current time to the start of the time window.
            window_length: The length of the time window.
            row_percentage: The fraction of rows to use when computing the statistics [0, 1.0].
        # Returns
            `FeatureMonitoringConfig`. The updated FeatureMonitoringConfig object.
        """
        # Setter is using the engine class to perform input validation.
        self.detection_window_config = {
            "window_config_type": (
                mwc.WindowConfigType.ROLLING_TIME
                if time_offset or window_length
                else mwc.WindowConfigType.ALL_TIME
            ),
            "time_offset": time_offset,
            "window_length": window_length,
            "row_percentage": row_percentage,
        }

        return self

    def with_reference_window(
        self,
        time_offset: Optional[str] = None,
        window_length: Optional[str] = None,
        row_percentage: Optional[float] = None,
    ) -> "FeatureMonitoringConfig":
        """Sets the reference window of data to compute statistics on.
        See also `with_reference_value(...)` and `with_reference_training_dataset(...)` for other reference options.
        !!! example
            ```python
            # Fetch your feature group or feature view
            fg = fs.get_feature_group(name="my_feature_group", version=1)
            # Setup feature monitoring and a detection window
            my_monitoring_config = fg.create_feature_monitoring(...).with_detection_window(...)
            # Statistics computed on a rolling time window, e.g. same day last week
            my_monitoring_config.with_reference_window(
                time_offset="1w",
                window_length="1d",
            ).compare_on(...).save()
            ```

        !!! warning "Provide a comparison configuration"
            You must provide a comparison configuration via `compare_on()` before saving the feature monitoring config.

        # Arguments
            time_offset: The time offset from the current time to the start of the time window.
            window_length: The length of the time window.
            row_percentage: The percentage of rows to use when computing the statistics. Defaults to 20%.
        # Returns
            `FeatureMonitoringConfig`. The updated FeatureMonitoringConfig object.
        """
        # Setter is using the engine class to perform input validation and build monitoring window config object.
        if self.detection_window_config is None:
            self.detection_window_config = {
                "window_config_type": mwc.WindowConfigType.ALL_TIME,
                "row_percentage": 1.0,
            }
        self.reference_window_config = {
            "time_offset": time_offset,
            "window_length": window_length,
            "row_percentage": row_percentage,
        }

        return self

    def with_reference_value(
        self,
        value: Optional[Union[float, int]] = None,
    ) -> "FeatureMonitoringConfig":
        """Sets the reference value to compare statistics with.
        See also `with_reference_window(...)` and `with_reference_training_dataset(...)` for other reference options.
        !!! example
            ```python
            # Fetch your feature group or feature view
            fg = fs.get_feature_group(name="my_feature_group", version=1)
            # Setup feature monitoring and a detection window
            my_monitoring_config = fg.create_feature_monitoring(...).with_detection_window(...)
            # Simplest reference window is a specific value
            my_monitoring_config.with_reference_value(
                value=0.0,
            ).compare_on(...).save()
            ```

        !!! warning "Provide a comparison configuration"
            You must provide a comparison configuration via `compare_on()` before saving the feature monitoring config.

        # Arguments
            value: A float value to use as reference.
        # Returns
            `FeatureMonitoringConfig`. The updated FeatureMonitoringConfig object.
        """
        self.reference_window_config = {
            "specific_value": value,
        }

        return self

    def with_reference_training_dataset(
        self,
        training_dataset_version: Optional[int] = None,
    ) -> "FeatureMonitoringConfig":
        """Sets the reference training dataset to compare statistics with.
        See also `with_reference_value(...)` and `with_reference_window(...)` for other reference options.
        !!! example
            ```python
            # Fetch your feature group or feature view
            fg = fs.get_feature_group(name="my_feature_group", version=1)
            # Setup feature monitoring and a detection window
            my_monitoring_config = fg.create_feature_monitoring(...).with_detection_window(...)
            # Only for feature views: Compare to the statistics computed for one of your training datasets
            # particularly useful if it has been used to train a model currently in production
            my_monitoring_config.with_reference_training_dataset(
                training_dataset_version=3,
            ).compare_on(...).save()
            ```

        !!! warning "Provide a comparison configuration"
            You must provide a comparison configuration via `compare_on()` before saving the feature monitoring config.

        # Arguments
            training_dataset_version: The version of the training dataset to use as reference.
        # Returns
            `FeatureMonitoringConfig`. The updated FeatureMonitoringConfig object.
        """
        self.reference_window_config = {
            "training_dataset_version": training_dataset_version,
        }

        return self

    def compare_on(
        self,
        metric: Optional[str],
        threshold: Optional[float],
        strict: Optional[bool] = False,
        relative: Optional[bool] = False,
    ) -> "FeatureMonitoringConfig":
        """Sets the statistics comparison criteria for feature monitoring with a reference window.
        !!! example
            ```python
            # Fetch your feature group or feature view
            fg = fs.get_feature_group(name="my_feature_group", version=1)
            # Setup feature monitoring, a detection window and a reference window
            my_monitoring_config = fg.create_feature_monitoring(
                ...
            ).with_detection_window(...).with_reference_window(...)
            # Choose a metric and set a threshold for the difference
            # e.g compare the relative mean of detection and reference window
            my_monitoring_config.compare_on(
                metric="mean",
                threshold=1.0,
                relative=True,
            ).save()
            ```

        !!! note
            Detection window and reference window/value/training_dataset must be set prior to comparison configuration.

        # Arguments
            metric: The metric to use for comparison. Different metric are available for different feature type.
            threshold: The threshold to apply to the difference to potentially trigger an alert.
            strict: Whether to use a strict comparison (e.g. > or <) or a non-strict comparison (e.g. >= or <=).
            relative: Whether to use a relative comparison (e.g. relative mean) or an absolute comparison (e.g. absolute mean).
        # Returns
            `FeatureMonitoringConfig`. The updated FeatureMonitoringConfig object.
        """
        # Setter is using the engine class to perform input validation.
        self.statistics_comparison_config = {
            "metric": metric,
            "threshold": threshold,
            "strict": strict,
            "relative": relative,
        }

        return self

    def save(self):
        """Saves the feature monitoring configuration.
        !!! example
            ```python
            # Fetch your feature group or feature view
            fg = fs.get_feature_group(name="my_feature_group", version=1)
            # Setup feature monitoring and a detection window
            my_monitoring_config = fg.create_statistics_monitoring(
                name="my_monitoring_config",
            ).save()
            ```
        # Returns
            `FeatureMonitoringConfig`. The saved FeatureMonitoringConfig object.
        """
        registered_config = self._feature_monitoring_config_engine.save(self)
        self.detection_window_config = registered_config._detection_window_config
        self.job_schedule = registered_config._job_schedule

        if (
            self._feature_monitoring_type
            != FeatureMonitoringType.STATISTICS_COMPUTATION
        ):
            self.reference_window_config = registered_config._reference_window_config
            self.statistics_comparison_config = (
                registered_config._statistics_comparison_config
            )

        self._job_name = registered_config._job_name
        self._id = registered_config._id
        return self

    def update(self):
        """Updates allowed fields of the saved feature monitoring configuration.
        !!! example
            ```python
            # Fetch your feature group or feature view
            fg = fs.get_feature_group(name="my_feature_group", version=1)
            # Fetch registered config by name
            my_monitoring_config = fg.get_feature_monitoring_configs(name="my_monitoring_config")
            # Update the percentage of rows to use when computing the statistics
            my_monitoring_config.detection_window.row_percentage = 10
            my_monitoring_config.update()
            ```
        # Returns
            `FeatureMonitoringConfig`. The updated FeatureMonitoringConfig object.
        """
        return self._feature_monitoring_config_engine.update(self)

    def run_job(self):
        """Trigger the feature monitoring job which computes and compares statistics on the detection and reference windows.
        !!! example
            ```python3
            # Fetch your feature group or feature view
            fg = fs.get_feature_group(name="my_feature_group", version=1)
            # Fetch registered config by name
            my_monitoring_config = fg.get_feature_monitoring_configs(name="my_monitoring_config")
            # Trigger the feature monitoring job once
            my_monitoring_config.run_job()
            ```

        !!! info
            The feature monitoring job will be triggered asynchronously and the method will return immediately.
            Calling this method does not affect the ongoing schedule.

        # Raises
            `FeatureStoreException`: If the feature monitoring config has not been saved.
        # Returns
            `Job`. A handle for the job computing the statistics.
        """
        if not self._id or not self._job_name:
            raise FeatureStoreException(
                "Feature monitoring config must be registered via `.save()` before computing statistics."
            )

        return self._feature_monitoring_config_engine.trigger_monitoring_job(
            job_name=self.job_name
        )

    def get_job(self):
        """Get the feature monitoring job which computes and compares statistics on the detection and reference windows.
        !!! example
            ```python3
            # Fetch registered config by name via feature group or feature view
            my_monitoring_config = fg.get_feature_monitoring_configs(name="my_monitoring_config")
            # Get the job which computes statistics on detection and reference window
            job = my_monitoring_config.get_job()
            # Print job history and ongoing executions
            job.executions
            ```
        # Raises
            `FeatureStoreException`: If the feature monitoring config has not been saved.
        # Returns
            `Job`. A handle for the job computing the statistics.
        """
        if not self._id or not self._job_name:
            raise FeatureStoreException(
                "Feature monitoring config must be registered via `.save()` before fetching"
                "the associated job."
            )

        return self._feature_monitoring_config_engine.get_monitoring_job(
            job_name=self.job_name
        )

    def delete(self):
        """Deletes the feature monitoring configuration.
        !!! example
            ```python
            # Fetch your feature group or feature view
            fg = fs.get_feature_group(name="my_feature_group", version=1)
            # Fetch registered config by name
            my_monitoring_config = fg.get_feature_monitoring_configs(name="my_monitoring_config")
            # Delete the feature monitoring config
            my_monitoring_config.delete()
            ```
        # Raises
            `FeatureStoreException`: If the feature monitoring config has not been saved.
        """
        if not self._id:
            raise FeatureStoreException(
                "Feature monitoring config must be registered via `.save()` before deleting."
            )

        self._feature_monitoring_config_engine.delete(config_id=self._id)

    def disable(self):
        """Disables the schedule of the feature monitoring job.
        !!! example
            ```python
            # Fetch your feature group or feature view
            fg = fs.get_feature_group(name="my_feature_group", version=1)
            # Fetch registered config by name
            my_monitoring_config = fg.get_feature_monitoring_configs(name="my_monitoring_config")
            # Disable the feature monitoring config
            my_monitoring_config.disable()
            ```
        # Raises
            `FeatureStoreException`: If the feature monitoring config has not been saved.
        """
        self._update_schedule(enabled=False)

    def enable(self):
        """Enables the schedule of the feature monitoring job.
        The scheduler can be configured via the `job_schedule` property.
        !!! example
            ```python
            # Fetch your feature group or feature view
            fg = fs.get_feature_group(name="my_feature_group", version=1)
            # Fetch registered config by name
            my_monitoring_config = fg.get_feature_monitoring_configs(name="my_monitoring_config")
            # Enable the feature monitoring config
            my_monitoring_config.enable()
            ```
        # Raises
            `FeatureStoreException`: If the feature monitoring config has not been saved.
        """
        self._update_schedule(enabled=True)

    def _update_schedule(self, enabled):
        if self._job_schedule is None:
            raise FeatureStoreException("No scheduler found for monitoring job")
        job_schedule = JobSchedule(
            id=self._job_schedule.id,
            start_date_time=self._job_schedule.start_date_time,
            cron_expression=self._job_schedule.cron_expression,
            end_time=self._job_schedule.end_date_time,
            enabled=enabled,
        )
        self.job_schedule = self._job_api.create_or_update_schedule_job(
            self._job_name, job_schedule.to_dict()
        )
        return self._job_schedule

    def get_history(
        self,
        start_time: Union[datetime, date, str, int, None] = None,
        end_time: Union[datetime, date, str, int, None] = None,
        with_statistics: bool = True,
    ) -> List["FeatureMonitoringResult"]:
        """
        Fetch the history of the computed statistics and comparison results for this configuration.
        !!! example
            ```python3
            # Fetch your feature group or feature view
            fg = fs.get_feature_group(name="my_feature_group", version=1)
            # Fetch registered config by name
            my_monitoring_config = fg.get_feature_monitoring_configs(name="my_monitoring_config")
            # Fetch the history of the computed statistics for this configuration
            history = my_monitoring_config.get_history(
                start_time="2021-01-01",
                end_time="2021-01-31",
            )
            ```
        # Args:
            start_time: The start time of the time range to fetch the history for.
            end_time: The end time of the time range to fetch the history for.
            with_statistics: Whether to include the computed statistics in the results.
        # Raises
            `FeatureStoreException`: If the feature monitoring config has not been saved.
        """
        if not self._id:
            raise FeatureStoreException(
                "Feature monitoring config must be registered via `.save()` before fetching"
                "the associated history."
            )
        return self._feature_monitoring_result_engine.fetch_all_feature_monitoring_results_by_config_id(
            config_id=self._id,
            start_time=start_time,
            end_time=end_time,
            with_statistics=with_statistics,
        )

    @property
    def id(self) -> Optional[int]:
        """Id of the feature monitoring configuration."""
        return self._id

    @property
    def feature_store_id(self) -> int:
        """Id of the Feature Store."""
        return self._feature_store_id

    @property
    def feature_group_id(self) -> Optional[int]:
        """Id of the Feature Group to which this feature monitoring configuration is attached."""
        return self._feature_group_id

    @property
    def feature_view_name(self) -> Optional[str]:
        """Name of the Feature View to which this feature monitoring configuration is attached."""
        return self._feature_view_name

    @property
    def feature_view_version(self) -> Optional[int]:
        """Version of the Feature View to which this feature monitoring configuration is attached."""
        return self._feature_view_version

    @property
    def feature_name(self) -> Optional[str]:
        """The name of the feature to monitor. If not set, all features of the
        Feature Group or Feature View are monitored, only available for scheduled statistics.

        !!! info "This property is read-only"

        """
        return self._feature_name

    @property
    def name(self) -> str:
        """The name of the feature monitoring config.
        A Feature Group or Feature View cannot have multiple feature monitoring configurations with the same name. The name of
        a feature monitoring configuration is limited to 63 characters.

        !!! info "This property is read-only once the feature monitoring configuration has been saved."
        """
        return self._name

    @name.setter
    def name(self, name: str):
        if hasattr(self, "_id"):
            raise AttributeError("The name of a registered config is read-only.")
        elif not isinstance(name, str):
            raise TypeError("name must be of type str")
        if len(name) > MAX_LENGTH_NAME:
            raise ValueError("name must be less than {MAX_LENGTH_NAME} characters.")
        self._name = name

    @property
    def description(self) -> Optional[str]:
        """Description of the feature monitoring configuration."""
        return self._description

    @description.setter
    def description(self, description: Optional[str]):
        if not isinstance(description, str) and description is not None:
            raise TypeError("description must be of type str")
        elif isinstance(description, str) and len(description) > MAX_LENGTH_DESCRIPTION:
            raise ValueError(
                "description must be less than {MAX_LENGTH_DESCRIPTION} characters"
            )
        self._description = description

    @property
    def job_name(self) -> Optional[str]:
        """Name of the feature monitoring job."""
        return self._job_name

    @property
    def enabled(self) -> bool:
        """Controls whether or not this config is spawning new feature monitoring jobs.
        This field belongs to the scheduler configuration but is made transparent to the user for convenience.
        """
        return self.job_schedule.enabled

    @enabled.setter
    def enabled(self, enabled: bool):
        """
        Controls whether or not this config is spawning new feature monitoring jobs.
        This field belongs to the scheduler configuration but is made transparent to the user for convenience.
        """
        self.job_schedule.enabled = enabled

    @property
    def feature_monitoring_type(self) -> Optional[str]:
        """The type of feature monitoring to perform. Used for internal validation.
        Options are:
            - STATISTICS_COMPUTATION if no reference window (and, therefore, comparison config) is provided
            - STATISTICS_COMPARISON if a reference window (and, therefore, comparison config) is provided.

        !!! info "This property is read-only."
        """
        return self._feature_monitoring_type

    @property
    def job_schedule(self) -> JobSchedule:
        """Schedule of the feature monitoring job.
        This field belongs to the job configuration but is made transparent to the user for convenience.
        """
        return self._job_schedule

    @job_schedule.setter
    def job_schedule(self, job_schedule: Union[JobSchedule, Dict[str, Any]]):
        if isinstance(job_schedule, JobSchedule):
            self._job_schedule = job_schedule
        elif isinstance(job_schedule, dict):
            self._job_schedule = JobSchedule(**job_schedule)
        else:
            raise TypeError("job_schedule must be of type JobScheduler, dict or None")

    @property
    def detection_window_config(self) -> mwc.MonitoringWindowConfig:
        """Configuration for the detection window."""
        return self._detection_window_config

    @detection_window_config.setter
    def detection_window_config(
        self,
        detection_window_config: Optional[
            Union[mwc.MonitoringWindowConfig, Dict[str, Any]]
        ],
    ):
        if isinstance(detection_window_config, mwc.MonitoringWindowConfig):
            self._detection_window_config = detection_window_config
        elif isinstance(detection_window_config, dict):
            self._detection_window_config = (
                self._monitoring_window_config_engine.build_monitoring_window_config(
                    **detection_window_config
                )
            )
        elif detection_window_config is None:
            self._detection_window_config = detection_window_config
        else:
            raise TypeError(
                "detection_window_config must be of type MonitoringWindowConfig, dict or None"
            )

    @property
    def reference_window_config(self) -> mwc.MonitoringWindowConfig:
        """Configuration for the reference window."""
        return self._reference_window_config

    @reference_window_config.setter
    def reference_window_config(
        self,
        reference_window_config: Optional[
            Union[mwc.MonitoringWindowConfig, Dict[str, Any]]
        ] = None,
    ):
        """Sets the reference window for monitoring."""
        # TODO: improve setter documentation
        if (
            self._feature_monitoring_type
            == FeatureMonitoringType.STATISTICS_COMPUTATION
            and reference_window_config is not None
        ):
            raise AttributeError(
                "reference_window_config is only available for feature monitoring"
                " not for scheduled statistics."
            )
        if isinstance(reference_window_config, mwc.MonitoringWindowConfig):
            self._reference_window_config = reference_window_config
        elif isinstance(reference_window_config, dict):
            self._reference_window_config = (
                self._monitoring_window_config_engine.build_monitoring_window_config(
                    **reference_window_config
                )
            )
        elif reference_window_config is None:
            self._reference_window_config = None
        else:
            raise TypeError(
                "reference_window_config must be of type MonitoringWindowConfig, dict or None"
            )

    @property
    def statistics_comparison_config(
        self,
    ) -> Optional[Dict[str, Any]]:
        """Configuration for the comparison of detection and reference statistics."""
        return self._statistics_comparison_config

    @statistics_comparison_config.setter
    def statistics_comparison_config(
        self,
        statistics_comparison_config: Optional[Dict[str, Any]] = None,
    ):
        if (
            self._feature_monitoring_type
            == FeatureMonitoringType.STATISTICS_COMPUTATION
            and statistics_comparison_config is not None
        ):
            raise AttributeError(
                "statistics_comparison_config is only available for feature monitoring"
                " not for scheduled statistics."
            )
        if isinstance(statistics_comparison_config, dict):
            self._statistics_comparison_config = self._feature_monitoring_config_engine.validate_statistics_comparison_config(
                **statistics_comparison_config
            )
        elif statistics_comparison_config is None:
            self._statistics_comparison_config = statistics_comparison_config
        else:
            raise TypeError("statistics_comparison_config must be of type dict or None")
